/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2003/06/27 09:18:38 $ $Revision: 1.4 $ */
#ifndef D2MGLOBAL_H
#define D2MGLOBAL_H

#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>

#ifndef M_PI
#define M_PI 3.1415926535897932384626433
#endif

#define strcasecmp(s1,s2) strcmp(s1,s2)

#ifndef max
#define max(a,b) (a<b?b:a)
#endif

#ifndef min
#define min(a,b) (a>b?b:a)
#endif

#ifndef fsgn
#define fsgn(a) (a==0.0?0.0:(a>0?1.0:-1.0))
#endif

typedef struct MGLtexdetail_ {
  GLenum *target;	/* the texture's target GL_TEXTURE_2D or 1D */
  GLint level;		/* level of detail, usually 0 */
  GLint components;	/* color components, usually 3 */
  GLsizei width,height;	/* ... */
  GLint border;		/* border width 0,1 */
  GLenum format;	/* format of pixel data, eg. GL_RGB */
  GLenum type;		/* type of pixel data, eg. GL_UNSIGNED_BYTE */
  unsigned char *data;	/* pixel data */
} MGLtexdetail;

typedef struct MGLtexture_ {
  int num;		/* Number of textures loaded */
  GLuint *texture;	/* the texture's name, an unsigned int */
  MGLtexdetail *detail;	/* details of the texture */
} MGLtexture;

struct lightstr {
  GLfloat colors[4][4];
};

typedef struct MGLquadratic_ {
  GLUquadricObj *q;	/* quadratics object needed for sphere */
  GLint		sl;	/* slices (longitude) usually =2*st */
  GLint		st;	/* stacks (latitude) */
} MGLquadratic;

typedef struct GLOBAL_ {
  int filetype;			/* can be derived */
  char *tmpdir;			/* def is /tmp */
  char *outfile;		/* def is movie.mpg */
  char *pattern;		/* def is IBBPBBPBBPBBPBBP */
  int pattlen;			/* to be derived */
  double framerate;		/* def is 25.0 */
  int framerepeat;		/* def is 1 */
  int quadres;			/* def is 32 */
  int quality;			/* def is 5 */
  int playdir;			/* def is 0 forward */
  double f;			/* def is 300, square coor multiplier */
  double x,y,z;			/* def is, 0,-5,-0.5 position of the origin */
  double dx,dy,dz;		/* def is 0,0,0 change of the latter */
  double phix,phiy,phiz;	/* def is, 0,-5,-0.5 position of the origin */
  double dphix,dphiy,dphiz;	/* def is 0,0,0 change of the latter */
  int startframe;		/* def is 0 */
  int endframe;			/* def is the last */
  int frameskip;		/* def is 1 */
  int width,height;		/* def is 300,300 */
  int lightnum;			/* def is 1 */
  struct lightstr *light;	/* should be reallocated if lightnum>1 */
  MGLtexture texture;		/* Texture parameters */
  double bg[4];			/* background color */
  MGLquadratic quad;		/* quadratic parameters */
  int i;			/* index for frames */
  void *os_buffer;		/* buffer of the gl image */
  char *png_name;		/* filename for pngs */
  char *png_templ;		/* template for pngs */
  int movie;			/* create movie (1) */
  int remove;			/* remove images (1) */
} GLOBAL;

GLOBAL gl;

typedef struct ANIMATION_ {
  int framepc;			/* position of the current frame */
  int framepossize;		/* size of framepos array */
  long *framepos;		/* position of the frames in the file */
  int maxframe;			/* total number of frames */
} ANIMATION;

ANIMATION anim;

int body_alloc;
#endif
