/* Copyright (C) 2002, by Sandor Fazekas and Janos Torok */
/* Last modified: $Date: 2004/11/02 11:12:33 $ $Revision: 1.30 $ */

#include "rpn.h"
#include "body.h"

#ifdef _WIN32
#define strncasecmp(s1,s2,n) strnicmp(s1,s2,n)
#endif

#ifndef min
#define min(a,b) (a>b?b:a)
#endif

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

#ifndef M_E
#define M_E 2.7182818284590452353602874713527
#endif

/* global variables */
const char RPN_openbracket[4]="({";
const char RPN_closebracket[4]=")}";
const char RPN_binaryoperator[RPN_BINNUM][RPN_StrLen]={ "^",
                "*", "/", "%",
                "+", "-", ">=", "==", "=", "<=", "!=", ">", "<",
                "&&", "&", "||", "|" };
const int  RPN_binoporder[RPN_BINNUM]={ 0,
                1, 1, 1,
                2, 2, 3, 3, 3, 3, 3, 3, 3,
                4, 4, 4, 4 };
const char RPN_unaryoperator[RPN_UONum][RPN_StrLen]={ "sin", "cos",
        "tg", "ctg", "asin", "acos", "atg", "actg", "sqrt",
        "sqr", "abs", "sgn", "exp", "log", "int", "norm",
	"var", "gbeg", "gnum" };
const char RPN_function[RPN_FNum][RPN_StrLen]={ "Rot", "Mirror", "Lincomb",
                        "If", "Vector", "Comp", "Projl" ,"Projp" };
const int  RPN_funcpar[RPN_FNum]={ 3, 2, 2,
                        3, 3, 2, 2, 2 };
const char RPN_funcparsep=',';
const char RPN_arrayvars[RPN_ArrVNum][RPN_StrLen]={ "i" };
const char RPN_partparam[RPN_StrLen]="P.";
const char RPN_arrayfunction[RPN_ArrFNum][RPN_StrLen]={ "Min", "Max",
                        "Av", "Sum", "Minp", "Maxp", "Foreach" };
const char RPN_number[12]="0123456789.";
const char RPN_sign[3]="+-";
const char RPN_vconst[RPN_VConstNum][RPN_StrLen]={ "i_x", "i_y",
                "i_z", "Nul", "White", "Black", "Red", "Green",
		"Blue", "Yellow", "Magenta", "Cyan", "Orange",
		"Brown", "Grey20", "Grey40", "Grey60", "Grey70",
		"Grey80", "Grey90", "Red2", "Red3", "Green2",
		"Green3", "Blue2", "Blue3", "Yellow2", "Yellow3",
		"Purple" };
const double RPN_vconstval[RPN_VConstNum][3]={ { 1.0, 0.0, 0.0 },
                { 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 }, { 0.0, 0.0, 0.0 },
		{ 1.0, 1.0, 1.0 }, { 0.0, 0.0, 0.0 }, { 1.0, 0.0, 0.0 },
		{ 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 }, { 1.0, 1.0, 0.0 },
		{ 1.0, 0.0, 1.0 }, { 0.0, 1.0, 1.0 }, { 1.0, 0.66, 0.0 },
		{ 0.5, 0.11, 0.11 }, { 0.2, 0.2, 0.2 }, { 0.4, 0.4, 0.4 },
		{ 0.6, 0.6, 0.6 }, { 0.7, 0.7, 0.7 }, { 0.8, 0.8, 0.8 },
		{ 0.9, 0.9, 0.9 }, { 0.75, 0.0, 0.0 }, { 0.5, 0.0, 0.0 },
		{ 0.0, 0.75, 0.0 }, { 0.0, 0.5, 0.0 }, { 0.0, 0.0, 0.75 },
		{ 0.0, 0.0, 0.5 }, { 0.75, 0.75, 0.0 }, { 0.5, 0.5, 0.0 },
		{ 0.6, 0.11, 1.0 } };
const char RPN_const[RPN_ConstNum][RPN_StrLen]={ "pi", "e",
	"wall", "ball", "cylinder", "disk", "force", "ellipsoid",
	"arrow", "cone", "triangle" };
const double RPN_constval[RPN_ConstNum]={ M_PI, M_E,
	1.0, 2.0, 3.0, 4.0, 6.0, 10.0, 12.0, 13.0,14.0 };
const double RP=180.0/M_PI;
const char RPN_parstr[CEF_PARNUM][CEF_SUBPARNUM][5]={
                { "num", "type", "hand", "" },
		{ "dx", "dy", "dz", "i1", "i2" },
                { "x", "y", "z", "" }, { "rx", "ry", "rz", "phi", "" },
		{ "cr", "cg", "cb", "ca", "" }, { "Z", "" },
		{ "vx", "vy", "vz", "" },
                { "fx", "fy", "fz", "" }, { "omx", "omy", "omz", "" },
                { "tx", "ty", "tz", "" }, { "tvx", "tvy", "tvz", "" },
                { "tfx", "tfy", "tfz", "" }, { "tox", "toy", "toz", "" },
		{ "life", "" }, { "mu", "im", "iI" }, { "tZ", "" },
		{ "dxvx", "dyvx", "dzvx" }, { "dxvy", "dyvy", "dzvy" }, 
		{ "dxvz", "dyvz", "dzvz" },
                { "ps", "t", "fpos", "" } /* this should be always the last */
                };
const char RPN_parvectstr[CEF_PARNUM][RPN_StrLen]={ "", "Dim", "Pos",
                "Orient", "Col", "", "Velo", "Force", "AngVelo",
		"Torque", "TrVelo", "TrForce", "TrAngVelo", "", "", "",
		"Dvx", "Dvy", "Dvz", "" };

/* Prototypes */
void Constraint_Error(char *);

/* for debug */
void printRPN(RPN *rpnroot)
{
  int a;
  RPN *rp;

  a=0;
  rp=rpnroot;
  while (rp!=NULL) {
    printf(
        "%2d rpn: %d t: %d r: %d v: %d ival: %d val: %lf (%.2lf,%.2lf,%.2lf)\n",
                a,rp->rpn,rp->type,rp->rank,rp->vect,rp->ival,rp->dval,
                rp->val.v[0],rp->val.v[1],rp->val.v[2]);
    rp=rp->next;
    a++;
  }
}

/* free allocated objects */
void RPN_Free(RPN **rpnroot)
{
  RPN *rp,*rp2;

  rp=*rpnroot;
  while (rp!=NULL) {
    rp2=rp;
    rp=rp->next;
    free(rp2);
  }
}

double fvnorm(vect *x,int vc)
{
  if (vc) return(sqrt(x->v[0]*x->v[0]+x->v[1]*x->v[1]+x->v[2]*x->v[2]));
  else return(x->v[0]);
}

int fvarray(vect *x,int pn,RPN *rp,int *v,int num,int arrayvar)
{
  int a,sv;
  vect vrv;
  double vtmp,mp;

  switch (rp->ival) {
    case 0 : /* Min */
      vrv.v[0]=x->v[0];
      vrv.v[1]=x->v[1];
      vrv.v[2]=x->v[2];
      sv=v[0];
      vtmp=fvnorm(&vrv,v[0]);
      for (a=1;a<pn;a++) {
        if (fvnorm(&x[a],v[a])<vtmp) {
          vrv.v[0]=x[a].v[0];
          vrv.v[1]=x[a].v[1];
          vrv.v[2]=x[a].v[2];
          sv=v[a];
          vtmp=fvnorm(&vrv,sv);
        }
      }
      if (arrayvar) {
        if ((!rp->arrnum) || (fvnorm(&rp->val,rp->vect)>vtmp)) {
          rp->val.v[0]=vrv.v[0];
          rp->val.v[1]=vrv.v[1];
          rp->val.v[2]=vrv.v[2];
          rp->vect=sv;
        }
      }
      rp->arrnum++;
      break;
    case 1 : /* Max */
      vrv.v[0]=x->v[0];
      vrv.v[1]=x->v[1];
      vrv.v[2]=x->v[2];
      sv=v[0];
      vtmp=fvnorm(&vrv,v[0]);
      for (a=1;a<pn;a++) {
        if (fvnorm(&x[a],v[a])>vtmp) {
          vrv.v[0]=x[a].v[0];
          vrv.v[1]=x[a].v[1];
          vrv.v[2]=x[a].v[2];
          sv=v[a];
          vtmp=fvnorm(&vrv,sv);
        }
      }
      if (arrayvar) {
        if ((!rp->arrnum) || (fvnorm(&rp->val,rp->vect)<vtmp)) {
          rp->val.v[0]=vrv.v[0];
          rp->val.v[1]=vrv.v[1];
          rp->val.v[2]=vrv.v[2];
          rp->vect=sv;
        }
      }
      rp->arrnum++;
      break;
    case 2 : /* Av */
      if (rp->arrnum) {
        vrv.v[0]=rp->val.v[0]*rp->arrnum;
        vrv.v[1]=rp->val.v[1]*rp->arrnum;
        vrv.v[2]=rp->val.v[2]*rp->arrnum;
      } else {
        vrv.v[0]=0.0;
        vrv.v[1]=0.0;
        vrv.v[2]=0.0;
      }
      sv=v[0];
      for (a=0;a<pn;a++) {
        vrv.v[0]+=x[a].v[0];
        vrv.v[1]+=x[a].v[1];
        vrv.v[2]+=x[a].v[2];
      }
      rp->arrnum++;
      if (arrayvar) {
        rp->val.v[0]=vrv.v[0]/rp->arrnum;
        rp->val.v[1]=vrv.v[1]/rp->arrnum;
        rp->val.v[2]=vrv.v[2]/rp->arrnum;
        rp->vect=sv;
      }
      break;
    case 3 : /* Sum */
      if (rp->arrnum) {
        vrv.v[0]=rp->val.v[0];
        vrv.v[1]=rp->val.v[1];
        vrv.v[2]=rp->val.v[2];
      } else {
        vrv.v[0]=0.0;
        vrv.v[1]=0.0;
        vrv.v[2]=0.0;
      }
      sv=v[0];
      for (a=0;a<pn;a++) {
        vrv.v[0]+=x[a].v[0];
        vrv.v[1]+=x[a].v[1];
        vrv.v[2]+=x[a].v[2];
      }
      if (arrayvar) {
        rp->val.v[0]=vrv.v[0];
        rp->val.v[1]=vrv.v[1];
        rp->val.v[2]=vrv.v[2];
        rp->vect=sv;
      }
      rp->arrnum++;
      break;
    case 4 : /* Minp */
      vrv.v[0]=x->v[0];
      vrv.v[1]=x->v[1];
      vrv.v[2]=x->v[2];
      mp=(double)num;
      sv=0;
      vtmp=fvnorm(&vrv,v[0]);
      for (a=1;a<pn;a++) {
        if (fvnorm(&x[a],v[a])<vtmp) {
          vrv.v[0]=x[a].v[0];
          vrv.v[1]=x[a].v[1];
          vrv.v[2]=x[a].v[2];
          mp=(double)num;
          vtmp=fvnorm(&vrv,sv);
        }
      }
      if (arrayvar) {
        if ((!rp->arrnum) || (rp->val.v[1]>vtmp)) {
          rp->val.v[0]=mp;
          rp->val.v[1]=vtmp;
        }
      }
      rp->arrnum++;
      vrv.v[0]=mp;
      break;
    case 5 : /* Maxp */
      vrv.v[0]=x->v[0];
      vrv.v[1]=x->v[1];
      vrv.v[2]=x->v[2];
      mp=(double)num;
      sv=0;
      vtmp=fvnorm(&vrv,v[0]);
      for (a=1;a<pn;a++) {
        if (fvnorm(&x[a],v[a])>vtmp) {
          vrv.v[0]=x[a].v[0];
          vrv.v[1]=x[a].v[1];
          vrv.v[2]=x[a].v[2];
          mp=(double)num;
          vtmp=fvnorm(&vrv,sv);
        }
      }
      if (arrayvar) {
        if ((!rp->arrnum) || (rp->val.v[1]<vtmp)) {
          rp->val.v[0]=mp;
          rp->val.v[1]=vtmp;
        }
      }
      rp->arrnum++;
      vrv.v[0]=mp;
      break;
    case 6 : /* Foreach */
      for (a=0;a<pn;a++) {
         if (fvnorm(&x[a],v[a])<=0.0) return(0);
      }
      vrv.v[0]=(double)num;
      vrv.v[1]=0.0;
      vrv.v[2]=0.0;
      *v=0;
      break;
  }
  x->v[0]=vrv.v[0];
  x->v[1]=vrv.v[1];
  x->v[2]=vrv.v[2];
  *v=sv;
  return(1);
}

void fvpartparam(vect *x,int fv,int v)
{
  int a;

  a=(int)x->v[0];
  if (a<0) a=0;
  if (a>=bg.body_num) a=bg.body_num-1;
  if (v) {
    x->v[0]=body[a].params[fv*CEF_SUBPARNUM];
    x->v[1]=body[a].params[fv*CEF_SUBPARNUM+1];
    x->v[2]=body[a].params[fv*CEF_SUBPARNUM+2];
  } else {
    x->v[0]=body[a].params[fv];
  }
}

void fvfunction(vect *x,vect *y,vect *z,int fv,int *vx,int *vy,int *vz)
{
  vect rv;
  double n,sp,csphi2;
  double xx,xy,xz,yy,yz,zz,sx,sy,sz;
  int sv,a;

  sv=0;
  switch (fv) {
    case 0: /* Rot(v,s,v) */
      if ((!*vx) || (*vy) || (!*vz)) return;
      sv=1;
      if (RPN_deg) y->v[0]/=RP;
      n=sqrt(x->v[0]*x->v[0]+x->v[1]*x->v[1]+x->v[2]*x->v[2]);
      csphi2=sin(y->v[0]*0.5);
      x->v[0]*=csphi2/n;
      x->v[1]*=csphi2/n;
      x->v[2]*=csphi2/n;
      csphi2=cos(y->v[0]*0.5);
      xx=x->v[0]*x->v[0];
      xy=x->v[0]*x->v[1];
      xz=x->v[0]*x->v[2];
      yy=x->v[1]*x->v[1];
      yz=x->v[1]*x->v[2];
      zz=x->v[2]*x->v[2];
      sx=csphi2*x->v[0];
      sy=csphi2*x->v[1];
      sz=csphi2*x->v[2];
      rv.v[0]=z->v[0]*(1.0-2.0*(yy+zz))+z->v[1]*2.0*(xy-sz)+z->v[2]*2.0*(xz+sy);
      rv.v[1]=z->v[0]*2.0*(xy+sz)+z->v[1]*(1.0-2.0*(xx+zz))+z->v[2]*2.0*(yz-sx);
      rv.v[2]=z->v[0]*2.0*(xz-sy)+z->v[1]*2.0*(yz+sx)+z->v[2]*(1.0-2.0*(xx+yy));
      break;
    case 1: /* Mirror(v,v) */
      if ((!*vx) || (!*vy)) return;
      sv=1;
      n=x->v[0]*x->v[0]+x->v[1]*x->v[1]+x->v[2]*x->v[2];
      sp=x->v[0]*y->v[0]+x->v[1]*y->v[1]+x->v[2]*y->v[2];
      rv.v[0]=y->v[0]-2*sp*x->v[0]/n;
      rv.v[1]=y->v[1]-2*sp*x->v[1]/n;
      rv.v[2]=y->v[2]-2*sp*x->v[2]/n;
      break;
    case 2: /* Lincomb(v,v) */
      if ((!*vx) || (!*vy)) return;
      sv=1;
      rv.v[0]=x->v[0]*y->v[0];
      rv.v[1]=x->v[1]*y->v[1];
      rv.v[2]=x->v[2]*y->v[2];
      break;
    case 3: /* If(s,s/v,s/v) */
      if (*vx) return;
      if ((!*vy) && (!*vz)) {
        sv=0;
        if (x->v[0]>0.0) {
          rv.v[0]=y->v[0];
        } else {
          rv.v[0]=z->v[0];
        }
      } else if ((*vy) && (*vz)) {
        sv=1;
        if (x->v[0]>0.0) {
          rv.v[0]=y->v[0];
          rv.v[1]=y->v[1];
          rv.v[2]=y->v[2];
        } else {
          rv.v[0]=z->v[0];
          rv.v[1]=z->v[1];
          rv.v[2]=z->v[2];
        }
      }
      break;
    case 4: /* Vector(s,s,s) */
      if ((*vx) || (*vy) || (*vz)) return;
      sv=1;
      rv.v[0]=x->v[0];
      rv.v[1]=y->v[0];
      rv.v[2]=z->v[0];
      break;
    case 5: /* Comp(v,s) */
      if ((!*vx) || (*vy)) return;
      sv=0;
      a=(int)y->v[0];
      if (a<0) a=0;
      if (a>2) a=2;
      rv.v[0]=x->v[a];
      break;
    case 6: /* Projl(v,v) */
      if ((!*vx) || (!*vy)) return;
      sv=1;
      n=x->v[0]*x->v[0]+x->v[1]*x->v[1]+x->v[2]*x->v[2];
      sp=x->v[0]*y->v[0]+x->v[1]*y->v[1]+x->v[2]*y->v[2];
      rv.v[0]=sp*x->v[0]/n;
      rv.v[1]=sp*x->v[1]/n;
      rv.v[2]=sp*x->v[2]/n;
      break;
    case 7: /* Projp(v,v) */
      if ((!*vx) || (!*vy)) return;
      sv=1;
      n=x->v[0]*x->v[0]+x->v[1]*x->v[1]+x->v[2]*x->v[2];
      sp=x->v[0]*y->v[0]+x->v[1]*y->v[1]+x->v[2]*y->v[2];
      rv.v[0]=y->v[0]-sp*x->v[0]/n;
      rv.v[1]=y->v[1]-sp*x->v[1]/n;
      rv.v[2]=y->v[2]-sp*x->v[2]/n;
      break;
  }
  x->v[0]=rv.v[0];
  x->v[1]=rv.v[1];
  x->v[2]=rv.v[2];
  *vx=sv;
}

void fvbinary(vect *x,vect *y,int fv,int *vx,int *vy)
{
  int vc;
  double valx,valy,valz,val;

  if (*vx) valx=sqrt(x->v[0]*x->v[0]+x->v[1]*x->v[1]+x->v[2]*x->v[2]);
  else valx=x->v[0];
  if (*vy) valy=sqrt(y->v[0]*y->v[0]+y->v[1]*y->v[1]+y->v[2]*y->v[2]);
  else valy=y->v[0];
  vc=0;

  switch (fv) {
    case 0: /* ^ */
      if (*vx+*vy==2) {
        /* vectorial product */
        vc=1;
        valx=x->v[1]*y->v[2]-y->v[1]*x->v[2];
        valy=x->v[2]*y->v[0]-y->v[2]*x->v[0];
        valz=x->v[0]*y->v[1]-y->v[0]*x->v[1];
        x->v[0]=valx;
        x->v[1]=valy;
        x->v[2]=valz;
      } else {
        val=pow(valx,valy);
      }
      break;
    case 1: /* * */
      if (*vx+*vy==2) {
        /* Scalar product */
        val=x->v[0]*y->v[0]+x->v[1]*y->v[1]+x->v[2]*y->v[2];
      } else if (*vx==1) {
        x->v[0]*=valy;
        x->v[1]*=valy;
        x->v[2]*=valy;
        vc=1;
      } else if (*vy==1) {
        x->v[0]=valx*y->v[0];
        x->v[1]=valx*y->v[1];
        x->v[2]=valx*y->v[2];
        vc=1;
      } else {
        val=valx*valy;
      }
      break;
    case 2: /* / */
      if (*vx) {
        x->v[0]/=valy;
        x->v[1]/=valy;
        x->v[2]/=valy;
        vc=1;
      } else {
        val=valx/valy;
      }
      break;
    case 3: /* % */
      val=fmod(valx,valy);
      break;
    case 4: /* + */
      /* strange but v+c=v, while c+v=c */
      if (*vx) {
        if (*vy) {
          x->v[0]+=y->v[0];
          x->v[1]+=y->v[1];
          x->v[2]+=y->v[2];
        } else {
          x->v[0]+=valy;
          x->v[1]+=valy;
          x->v[2]+=valy;
        }
        vc=1;
      } else {
        val=valx+valy;
      }
      break;
    case 5: /* - */
      if (*vx) {
        if (*vy) {
          x->v[0]-=y->v[0];
          x->v[1]-=y->v[1];
          x->v[2]-=y->v[2];
        } else {
          x->v[0]-=valy;
          x->v[1]-=valy;
          x->v[2]-=valy;
        }
        vc=1;
      } else {
        val=valx-valy;
      }
      break;
    case 6: /* >= */
      val=(double)(valx>=valy);
      break;
    case 7: /* == */
    case 8: /* = */
      val=(double)(valx==valy);
      break;
    case 9: /* <= */
      val=(double)(valx<=valy);
      break;
    case 10: /* != */
      val=(double)(valx!=valy);
      break;
    case 11: /* > */
      val=(double)(valx>valy);
      break;
    case 12: /* < */
      val=(double)(valx<valy);
      break;
    case 13: /* && */
    case 14: /* & */
      val=(double)((valx>0.0)*(valy>0.0));
      break;
    case 15: /* || */
    case 16: /* | */
      val=(double)((valx>0.0)+(valy>0.0));
      break;
  }
  if (!vc) {
    x->v[0]=val;
  }
  *vx=vc;
}

void fvunary(vect *x,int fv,int *v)
{
  int vc,i;
  double deg;
  double val;

  if (*v) val=sqrt(x->v[0]*x->v[0]+x->v[1]*x->v[1]+x->v[2]*x->v[2]);
  else val=x->v[0];
  if (RPN_deg) deg=val/RP;
  else deg=val;
  vc=0;

  switch (fv) {
    case 0: /* sin */
      val=sin(deg);
      break;
    case 1: /* cos */
      val=cos(deg);
      break;
    case 2: /* tan */
      val=tan(deg);
      break;
    case 3: /* ctg */
      val=1.0/tan(deg);
      break;
    case 4: /* asin */
      if (fabs(deg)>1) val=0.0;
      else {
        if (RPN_deg) val=asin(val)*RP;
        else val=asin(val);
      }
      break;
    case 5: /* acos */
      if (fabs(deg)>1) val=0.0;
      else {
        if (RPN_deg) val=acos(val)*RP;
        else val=acos(val);
      }
      break;
    case 6: /* atg */
      if (RPN_deg) val=atan(val)*RP;
      else val=atan(val);
      break;
    case 7:  /* actg */
      if (RPN_deg) val=atan(1.0/val)*RP;
      else val=atan(1.0/val);
      break;
    case 8: /* sqrt */
      val=sqrt(fabs(val));
      break;
    case 9: /* sqr */
      val*=val;
      break;
    case 10: /* abs */
      val=fabs(val);
      break;
    case 11: /* sgn */
      if (val!=0.0) {
        val=(val>0 ? 1.0 : -1.0);
      }
      break;
    case 12: /* exp */
      val=exp(val);
      break;
    case 13: /* log */
      if (val<=0.0) val=0.0;
      else val=log(val);
      break;
    case 14: /* int */
      val=floor(val);
      break;
    case 15: /* norm */
      if (*v) {
        vc=1;
        x->v[0]/=val;
        x->v[1]/=val;
        x->v[2]/=val;
      } else {
        if (val!=0.0) {
          val=(val>0 ? 1.0 : -1.0);
        }
      }
      break;
    case 16: /* var */
#ifdef RPN_var
      i=(int)val;
      if (i<0) i=0;
      if (i>=cvg.var_num) i=cvg.var_num-1;
      val=x->v[0]=cvg.var[i].v[0];
      x->v[1]=cvg.var[i].v[1];
      x->v[2]=cvg.var[i].v[2];
      vc=cvg.vc[i];
#else
      x->v[0]=0.0;
      x->v[1]=0.0;
      x->v[2]=0.0;
#endif
      break;
    case 17: /* gbeg */
      val=(double)bg.gbeg[(int)(val)-1];
      break;
    case 18: /* gnum */
      val=(double)bg.gnum[(int)(val)-1];
      break;
  }
  if (!vc) {
    x->v[0]=val;
  }
  *v=vc;
}

int RPN_Evaluate(RPN *rpnroot,double *var,double **global,double *scalar,
                                vect *vector,int *sv,int mode,int n)
{
  vect *stack;
  int sp,*stackvc,*stackarray,spa,containsi;
  RPN *rp;

  /* count the number of objects */
  rp=rpnroot;
  if (rp==NULL) {
    if (scalar!=NULL) *scalar=1.0;
    return(1);
  }
  spa=sp=0;
  while (rp!=NULL) {
    if ((rp->type==RpnT_Var) || (rp->type==RpnT_Const) ||
        (rp->type==RpnT_ArrayVar) || (rp->type==RpnT_GlobalVar2)) sp++;
    if (rp->type==RpnT_ArrayFunc) spa++;
    rp=rp->next;
  }
  stack=(vect *)calloc(sp,sizeof(vect));
  stackvc=(int *)calloc(sp,sizeof(int));
  stackarray=(int *)calloc(spa,sizeof(int));
  sp=-1;
  spa=-1;
  rp=rpnroot;
  while (rp!=NULL) {
    switch (rp->type) {
      case RpnT_Const:
        sp++;
        if (rp->vect) {
          stack[sp].v[0]=rp->val.v[0];
          stack[sp].v[1]=rp->val.v[1];
          stack[sp].v[2]=rp->val.v[2];
        } else {
          stack[sp].v[0]=rp->dval;
        }
        stackvc[sp]=rp->vect;
        break;
      case RpnT_Var:
        sp++;
        if (rp->vect) {
          stack[sp].v[0]=var[rp->ival*CEF_SUBPARNUM];
          stack[sp].v[1]=var[rp->ival*CEF_SUBPARNUM+1];
          stack[sp].v[2]=var[rp->ival*CEF_SUBPARNUM+2];
        } else {
          stack[sp].v[0]=var[rp->ival];
        }
        stackvc[sp]=rp->vect;
        break;
      case RpnT_ArrayVar:
        sp++;
        stack[sp].v[0]=(double)n;
        stackvc[sp]=0;
        containsi=1;
        break;
      case RpnT_Unary:
        if (sp>=0) {
          fvunary(&stack[sp],rp->ival,&stackvc[sp]);
        } else {
          /* error */
          Constraint_Error("Function without operand");
          free(stack);
          free(stackvc);
          free(stackarray);
          return(0);
        }
        break;
      case RpnT_Binary:
        if (sp>0) {
          sp--;
          fvbinary(&stack[sp],&stack[sp+1],rp->ival,
                                &stackvc[sp],&stackvc[sp+1]);
        } else {
          /* error */
          free(stack);
          free(stackvc);
          free(stackarray);
          return(0);
        }
        break;
      case RpnT_Function:
        if (sp>=RPN_funcpar[rp->ival]-1) {
          sp-=RPN_funcpar[rp->ival]-1;
          fvfunction(&stack[sp],&stack[sp+1],&stack[sp+2],rp->ival,
                        &stackvc[sp],&stackvc[sp+1],&stackvc[sp+2]);
        } else {
          /* error */
          Constraint_Error("Function without enough operand");
          free(stack);
          free(stackvc);
          free(stackarray);
          return(0);
        }
        break;
      case RpnT_ArrayEnd:
        spa++;
        stackarray[spa]=sp+1;
        containsi=0;
        break;
      case RpnT_ArrayFunc:
        if (sp>=0) {
          if ((containsi==0) && (mode<3)) {
            fvarray(&stack[stackarray[spa]],sp-stackarray[spa]+1,
                                rp,&stackvc[stackarray[spa]],n,0);
          } else {
            if (mode&2) {
              if (!fvarray(&stack[stackarray[spa]],sp-stackarray[spa]+1,
                        rp,&stackvc[stackarray[spa]],n,1)) {
                if (scalar!=NULL) *scalar=0.0;
                if (vector!=NULL) {
                  vector->v[0]=0.0;
                  vector->v[1]=0.0;
                  vector->v[2]=0.0;
                }
                free(stack);
                free(stackvc);
                free(stackarray);
                return(0);
              }
            } else {
              sp=stackarray[spa];
              stack[sp].v[0]=rp->val.v[0];
              stack[sp].v[1]=rp->val.v[1];
              stack[sp].v[2]=rp->val.v[2];
              stackvc[sp]=rp->vect;
              rp->arrnum=0;
            }
          }
          sp=stackarray[spa];
        } else {
          /* error */
          Constraint_Error("Function without operand");
          free(stack);
          free(stackvc);
          free(stackarray);
          return(0);
        }
        break;
      case RpnT_PartParam:
        if (sp>=0) {
          fvpartparam(&stack[sp],rp->ival,rp->vect);
          stackvc[sp]=rp->vect;
        } else {
          /* error */
          Constraint_Error("Function without operand");
          free(stack);
          free(stackvc);
          free(stackarray);
          return(0);
        }
        break;
      case RpnT_GlobalVar2:
        sp++;
	stack[sp].v[0]=*(global[rp->ival]);
        stackvc[sp]=0;
	break;
    }
    rp=rp->next;
  }
  if (sv!=NULL) *sv=stackvc[0];
  if (stackvc[0]) {
    if (vector!=NULL) {
      vector->v[0]=stack[0].v[0];
      vector->v[1]=stack[0].v[1];
      vector->v[2]=stack[0].v[2];
    }
    if (scalar!=NULL) *scalar=sqrt(stack[0].v[0]*stack[0].v[0]+
          stack[0].v[1]*stack[0].v[1]+stack[0].v[2]*stack[0].v[2]);
  } else {
    if (scalar!=NULL) *scalar=stack[0].v[0];
    if (vector!=NULL) {
      vector->v[0]=stack[0].v[0];
      vector->v[1]=0.0;
      vector->v[2]=0.0;
    }
  }
  free(stack);
  free(stackvc);
  free(stackarray);
  return(1);
}

/* move object after already converted objects to the right */
void GoOver(RPN *rp,RPN *prp,RPN **rpnroot)
{
  RPN *tmp,*prev;

  if (rp->next==NULL) return;
  if (rp->next->rpn==0) return;
  if (rp==(*rpnroot)) {
    *rpnroot=(*rpnroot)->next;
  } else {
    prp->next=rp->next;
  }
  prev=rp;tmp=rp->next;
  while ((tmp!=NULL) && (tmp->rpn)) {
    prev=tmp;tmp=tmp->next;
  }
  rp->next=prev->next;
  prev->next=rp;
}

/* operator order: ^ (* / %) (+ -) (= < >) (& |) */
int toRPN(RPN **rpnroot)
{
  int a,rank;
  RPN *rp,*prev;

  rank=0;
  rp=(*rpnroot);
  while (rp!=NULL) {
    if (rp->rank>rank) rank=rp->rank;
    rp=rp->next;
  }
  while (rank>=0) {
    /* main loop */
    prev=rp=(*rpnroot);
    while (rp!=NULL) {
    /* Unary operators are before binary ones */
      if ((rp->rank==rank) && ((rp->type==RpnT_Unary) ||
            (rp->type==RpnT_Function) || (rp->type==RpnT_ArrayFunc) ||
            (rp->type==RpnT_PartParam)) &&
          (!rp->rpn)) {
        rp->rpn=1;
        GoOver(rp,prev,rpnroot);
      }
      prev=rp;
      rp=rp->next;
    }
    /* Finally binary operators */
    for (a=0;a<=RPN_binoporder[RPN_BINNUM-1];a++) {
      prev=rp=(*rpnroot);
      while (rp!=NULL) {
        if ((rp->rank==rank) && (rp->type==RpnT_Binary) &&
            (RPN_binoporder[rp->ival]==a) && (!rp->rpn)) {
          rp->rpn=1;
          GoOver(rp,prev,rpnroot);
        }
        prev=rp;
        rp=rp->next;
      }
    }
    /* Section end */
    rp=(*rpnroot);
    while (rp!=NULL) {
      if ((rp->rank==rank) && (rp->type==RpnT_SecEnd)) {
        rp->rpn=1;
      }
      rp=rp->next;
    }
    rank--;
  }
  return(1);
}

double tonumber(char **sp,char **error)
{
  double sgn,val,d;

  *error=NULL;
  val=0.0;
  sgn=1.0;
  if (strchr(RPN_sign,**sp)) {
    sgn=(**sp=='-'?-1.0:1.0);
    (*sp)++;
  }
  while (isdigit(**sp)) {
    val=val*10.0+ (**sp-'0');
    (*sp)++;
  }
  if (**sp=='.') {
    (*sp)++;
    d=0.1;
    while (isdigit(**sp)) {
      val=val+d* (**sp-'0');
      d/=10.0;
      (*sp)++;
    }
  }
  if (tolower(**sp)=='e') {
    if (val==0.0) {
      *error=strdup("Invalid number!");
      return(0.0);
    }
    (*sp)++;
    val*=pow(10.0,tonumber(sp,error));
  }
  return(val*sgn);
}

int Find_Function(char **s,int limit,int limit2,
        RPN_strtype array,
        const char array2[CEF_PARNUM][CEF_SUBPARNUM][5],
        int dim,int cas,int alnum)
{
  int a,b;
  int (*scmp)(const char *,const char *,size_t);

  if (cas) {
    scmp=strncmp;
  } else {
#ifdef _WIN32
    scmp=_strnicmp;
#else
    scmp=strncasecmp;
#endif
  }
  if (dim) {
    for (a=0;a<limit;a++) {
      for (b=0;b<limit2;b++) {
        if ((strlen(array2[a][b])>0) &&
            (scmp(array2[a][b],*s,strlen(array2[a][b]))==0) &&
            (!isalnum(*(*s+min(strlen(array2[a][b]),strlen(*s)))))) {
          *s+=strlen(array2[a][b]);
          return(a*limit2+b);
        }
      }
    }
  } else {
    for (a=0;a<limit;a++) {
      if ((strlen(array[a])>0) && (scmp(array[a],*s,strlen(array[a]))==0) &&
          ((!isalnum(*(*s+min(strlen(array[a]),strlen(*s))))) || (alnum))) {
        *s+=strlen(array[a]);
        return(a);
      }
    }
  }
  return(-1);
}

int RPN_decode(char *s,RPN **rpnroot,char **error,int vart,int *arrayvar,
    		int nglob,RPN_strtype globstr)
{
  int a,rank,first,l;
  int begin;            /* beginning of a subexpr needed for sign */
  char *sp,ch,*spp;
  char star[]="*";
  RPN *rp;

  rank=0;
  begin=first=1;
  sp=s;
  while (*sp) {
    while (isspace(*sp)) sp++;
    if (*sp) {
      if (strchr(RPN_openbracket,*sp)) {
        rank++;
        begin=1;
        sp++;
      } else {
        if (first) {
          first=0;
          rp=*rpnroot=(RPN *)malloc(sizeof(RPN));
        } else {
          rp->next=(RPN *)malloc(sizeof(RPN));
          rp=rp->next;
        }
        rp->next=NULL;
        rp->rpn=0;
        rp->vect=0;
        rp->rank=rank;
        if (strchr(RPN_closebracket,*sp)) {
          begin=0;
          sp++;
          rp->rank=rank;
          rp->type=RpnT_SecEnd;
          rank--;
          if (rank<0) {
            *error=strdup("Too many closing bracket");
            RPN_Free(rpnroot);
            return(0);
          }
        } else if (*sp==RPN_funcparsep) {
          sp++;
          begin=1;
          rp->type=RpnT_SecEnd;
        } else if ((begin) && (strchr(RPN_sign,*sp)) && 
            (!strchr(RPN_number,*(sp+1)))) {
          /* check sign before an object */
          rp->dval=(*sp=='-' ? -1.0 : 1.0);
          rp->type=RpnT_Const;
          rp->rpn=1;
          rp->next=(RPN *)malloc(sizeof(RPN));
          rp=rp->next;
          rp->rpn=0;
          rp->type=RpnT_Binary;
          spp=star;
          rp->ival=Find_Function(&spp,RPN_BINNUM,0,
                                RPN_binaryoperator,NULL,0,0,1);
          rp->next=NULL;
          rp->rank=rank;
          sp++;
        } else if ((strchr(RPN_number,*sp)) || ((begin) &&
            (strchr(RPN_sign,*sp)) && (strchr(RPN_number,*(sp+1))))) {
          /* check number */
          begin=0;
          rp->dval=tonumber(&sp,error);
          if (*error!=NULL) {
            RPN_Free(rpnroot);
            return(0);
          }
          rp->type=RpnT_Const;
          rp->rpn=1;
        } else if ((a=Find_Function(&sp,RPN_BINNUM,0,
                                RPN_binaryoperator,NULL,0,0,1))>=0) {
          /* check binary operator */
          begin=1;
          rp->ival=a;
          rp->type=RpnT_Binary;
        } else if ((a=Find_Function(&sp,RPN_ConstNum,0,
                                RPN_const,NULL,0,0,0))>=0) {
          /* check constant */
          rp->type=RpnT_Const;
          rp->dval=RPN_constval[a];
          rp->rpn=1;
          begin=0;
        } else if ((a=Find_Function(&sp,RPN_VConstNum,0,
                                RPN_vconst,NULL,0,0,0))>=0) {
          /* check vector constant */
          rp->type=RpnT_Const;
          rp->val.v[0]=RPN_vconstval[a][0];
          rp->val.v[1]=RPN_vconstval[a][1];
          rp->val.v[2]=RPN_vconstval[a][2];
          rp->vect=1;
          rp->rpn=1;
          begin=0;
        } else if ((a=Find_Function(&sp,RPN_UONum,0,
                                RPN_unaryoperator,NULL,0,0,0))>=0) {
          /* check unary operator */
          rp->type=RpnT_Unary;
          rp->ival=a;
          begin=1;
        } else if (strncasecmp(sp,RPN_partparam,strlen(RPN_partparam))==0) {
          /* check particle parameter */
          sp+=2;
          rp->type=RpnT_PartParam;
          if ((a=Find_Function(&sp,CEF_PARNUM,0,
                                RPN_parvectstr,NULL,0,0,0))>=0) {
            rp->ival=a;
            rp->vect=1;
          } else if ((a=Find_Function(&sp,CEF_PARNUM,CEF_SUBPARNUM,
                                NULL,RPN_parstr,1,1,0))>=0) {
            rp->ival=a;
            rp->vect=0;
          } else {
            *error=strdup("Unkown particle parameter");
            RPN_Free(rpnroot);
            return(0);
          }
          begin=1;
        } else if ((a=Find_Function(&sp,RPN_FNum,0,
                                RPN_function,NULL,0,1,0))>=0) {
          /* check function */
          rp->type=RpnT_Function;
          rp->ival=a;
          begin=1;
        } else if ((a=Find_Function(&sp,RPN_ArrFNum,0,
                                RPN_arrayfunction,NULL,0,1,0))>=0) {
          /* check array function */
          rp->type=RpnT_ArrayFunc;
          rp->ival=a;
          rp->arrnum=0;
          begin=1;
          rp->next=(RPN *)malloc(sizeof(RPN));
          rp=rp->next;
          rp->rpn=1;
          rp->rank=rank;
          rp->type=RpnT_ArrayEnd;
          RPN_arrayfunc++;
        } else if ((a=Find_Function(&sp,RPN_ArrVNum,0,
                                RPN_arrayvars,NULL,0,1,0))>=0) {
          /* array function variable can only be used in array functions */
          if (arrayvar!=NULL) (*arrayvar)++;
          rp->type=RpnT_ArrayVar;
          rp->ival=a;
          rp->vect=0;
          rp->rpn=1;
          begin=0;
        } else if ((a=Find_Function(&sp,CEF_PARNUM,0,
                                RPN_parvectstr,NULL,0,0,0))>=0) {
          /* vector variable */
          rp->type=RpnT_Var;
          rp->ival=a;
          rp->vect=1;
          rp->rpn=1;
          begin=0;
        } else if ((a=Find_Function(&sp,CEF_PARNUM,CEF_SUBPARNUM,
                                NULL,RPN_parstr,1,1,0))>=0) {
          /* variable */
          rp->type=RpnT_Var;
          rp->ival=a;
          rp->rpn=1;
          begin=0;
        } else if ((a=Find_Function(&sp,nglob,0,
			globstr,NULL,0,1,1))>=0) {
          /* global variable */
          rp->type=RpnT_GlobalVar2;
          rp->ival=a;
          rp->rpn=1;
          rp->vect=0;
          begin=0;
        } else {
          l=0;
          while (isalnum(*(sp+l))) l++;
          *error=(char *)calloc(strlen("Unkown object: ")+
                    l+10,sizeof(char));
          ch=*(sp+l);
          *(sp+l)='\0';
          sprintf(*error,"Unkown object: %s",sp);
          *(sp+l)=ch;
          RPN_Free(rpnroot);
          return(0);
        }
      }
    }
  }
  if (rank!=0) {
    *error=strdup("Too many opening bracket");
    RPN_Free(rpnroot);
    return(0);
  }
  return(toRPN(rpnroot));
}
